#include <iostream>
#include <sstream>

#include <windows.h>

#include "pxcsensemanager.h"

#include <opencv2\opencv.hpp>

class RealSenseAsenseManager
{
public:

    ~RealSenseAsenseManager()
    {
        if ( scanner != nullptr ){
            scanner->Release();
            scanner = nullptr;
        }

        if ( senseManager != nullptr ){
            senseManager->Release();
            senseManager = nullptr;
        }
    }

    void initilize()
    {
        // SenseManager𐶐
        senseManager = PXCSenseManager::CreateInstance();
        if ( senseManager == 0 ) {
            throw std::runtime_error( "SenseManager̐Ɏs܂" );
        }

        // 3DXLLɂ
        pxcStatus sts = senseManager->Enable3DScan();
        if ( sts < PXC_STATUS_NO_ERROR ) {
            throw std::runtime_error( "3DXL̗LɎs܂" );
        }

        // pCvC
        sts = senseManager->Init();
        if ( sts < PXC_STATUS_NO_ERROR ) {
            throw std::runtime_error( "pCvC̏Ɏs܂" );
        }

        // ~[\ɂ
        senseManager->QueryCaptureManager()->QueryDevice()->SetMirrorMode(
            PXCCapture::Device::MirrorMode::MIRROR_MODE_HORIZONTAL );

        // 3DXL̏
        initialize3dScan();
    }

    void run()
    {
        // C[v
        while ( 1 ) {
            // t[f[^XV
            updateFrame();

            // \
            auto ret = showImage();
            if ( !ret ){
                break;
            }
        }
    }

private:

    void initialize3dScan()
    {
        // XLi[擾
        scanner = senseManager->Query3DScan();
        if ( scanner == nullptr ){
            throw std::runtime_error( "XLi[̎擾Ɏs܂" );
        }

        // ^[QbgIvV̐ݒ
        setTargetingOption(
            PXC3DScan::TargetingOption::NO_TARGETING_OPTIONS );

        // XL[h̐ݒ
        setScanMode( PXC3DScan::Mode::TARGETING );

        // f쐬IvV̕\
        showReconstructionOption();
        showModelFormat();
    }

    void updateFrame()
    {
        // t[擾
        pxcStatus sts = senseManager->AcquireFrame( false );
        if ( sts < PXC_STATUS_NO_ERROR ) {
            return;
        }

        // t[f[^擾
        updateColorImage( scanner->AcquirePreviewImage() );

        // t[
        senseManager->ReleaseFrame();
    }

    // J[摜XV
    void updateColorImage( PXCImage* colorFrame )
    {
        if ( colorFrame == 0 ){
            return;
        }
            
        PXCImage::ImageInfo info = colorFrame->QueryInfo();

        // f[^擾
        PXCImage::ImageData data;
        pxcStatus sts = colorFrame->AcquireAccess( PXCImage::Access::ACCESS_READ,
            PXCImage::PixelFormat::PIXEL_FORMAT_RGB24, &data );
        if ( sts < PXC_STATUS_NO_ERROR ) {
            throw std::runtime_error("J[摜̎擾Ɏs");
        }

        // f[^Rs[
        colorImage = cv::Mat( info.height, info.width, CV_8UC3 );
        memcpy( colorImage.data, data.planes[0], info.height * info.width * 3 );

        // f[^
        colorFrame->ReleaseAccess( &data );
    }

    // 摜\
    bool showImage()
    {
        // \
        cv::imshow( "Color Image", colorImage );

        int c = cv::waitKey( 10 );
        if ( (c == 27) || (c == 'q') || (c == 'Q') ){
            // ESC|q|Q for Exit
            return false;
        }
        else if ( c == 't' ) {
            // ^[QbgIvVύX
            auto option = scanner->QueryTargetingOptions();
            if ( option == PXC3DScan::TargetingOption::NO_TARGETING_OPTIONS ){
                setTargetingOption(
                    PXC3DScan::TargetingOption::OBJECT_ON_PLANAR_SURFACE_DETECTION );
            }
            else{
                setTargetingOption(
                    PXC3DScan::TargetingOption::NO_TARGETING_OPTIONS );
            }
        }
        else if ( c == 's' ) {
            // XL[hύX
            auto scanMode = scanner->QueryMode();
            if ( scanMode == PXC3DScan::Mode::TARGETING ){
                setScanMode( PXC3DScan::Mode::SCANNING );
            }
            else{
                setScanMode( PXC3DScan::Mode::TARGETING );
            }
        }
        else if ( c == 'o' ){
            // f쐬IvVύX
            changeReconstructionOption();
        }
        else if ( c == 'f' ){
            // f쐬tH[}bgύX
            changeModelFormat();
        }
        else if ( c == 'r' ){
            // f쐬
            reconstruct();
        }

        return true;
    }

    // ^[QbgIvVݒ肷
    void setTargetingOption( PXC3DScan::TargetingOption targetingOption )
    {
        std::cout << "setTargetingOption " << targetingOption << std::endl;
        auto sts = scanner->SetTargetingOptions( targetingOption );
        if ( sts < PXC_STATUS_NO_ERROR ){
            throw std::runtime_error( "^[QbgIvV̐ݒɎs܂" );
        }
    }

    // XL[hݒ肷
    void setScanMode( PXC3DScan::Mode scanMode )
    {
        std::cout << "setScanMode " << scanMode << std::endl;
        auto sts = scanner->SetMode( scanMode );
        if ( sts < PXC_STATUS_NO_ERROR ){
            throw std::runtime_error( "XL[h̐ݒɎs܂" );
        }
    }

    // f쐬IvVύX
    void changeReconstructionOption()
    {
        if ( reconstructionOption ==
            PXC3DScan::ReconstructionOption::NO_RECONSTRUCTION_OPTIONS ){
            reconstructionOption = PXC3DScan::ReconstructionOption::SOLIDIFICATION;
        }
        else {
            reconstructionOption = PXC3DScan::ReconstructionOption::NO_RECONSTRUCTION_OPTIONS;
        }

        showReconstructionOption();
    }

    // f쐬IvV\
    void showReconstructionOption()
    {
        auto option = (reconstructionOption ==
            PXC3DScan::ReconstructionOption::NO_RECONSTRUCTION_OPTIONS) ?
                "NO_RECONSTRUCTION_OPTIONS" : "SOLIDIFICATION";

        std::cout << "Reconstruction Option : " << option << std::endl;
    }

    // ftH[}bgύX
    void changeModelFormat()
    {
        if ( fileFormat == PXC3DScan::FileFormat::OBJ ){
            fileFormat = PXC3DScan::FileFormat::STL;
        }
        else if ( fileFormat == PXC3DScan::FileFormat::STL ){
            fileFormat = PXC3DScan::FileFormat::PLY;
        }
        else {
            fileFormat = PXC3DScan::FileFormat::OBJ;
        }

        showModelFormat();
    }

    // ftH[}bg\
    void showModelFormat()
    {
        std::wcout << L"Model Format          : " <<
            PXC3DScan::FileFormatToString( fileFormat ) << std::endl;
    }


    // f쐬
    void reconstruct()
    {
        // XLȊO̓f쐬Ȃ
        auto scanMode = scanner->QueryMode();
        if ( scanMode != PXC3DScan::Mode::SCANNING ){
            return;
        }

        // t@C쐬
        WCHAR fileTitle[MAX_PATH];
        GetTimeFormatEx( 0, 0, 0, L"hhmmss", fileTitle, _countof( fileTitle ) );

        std::wstringstream ss;
        ss << L"model-" << fileTitle << L"." << PXC3DScan::FileFormatToString( fileFormat );

        std::wcout << L"create " << ss.str() << "...";


        // 3Df쐬
        scanner->Reconstruct( fileFormat, ss.str().c_str(), reconstructionOption );

        std::cout << "done." << std::endl;
    }

private:

    cv::Mat colorImage;
    PXCSenseManager *senseManager = nullptr;
    PXC3DScan* scanner = nullptr;

    PXC3DScan::ReconstructionOption reconstructionOption =
        PXC3DScan::ReconstructionOption::NO_RECONSTRUCTION_OPTIONS;
    PXC3DScan::FileFormat fileFormat = PXC3DScan::FileFormat::OBJ;
};

void main()
{
    try{
        RealSenseAsenseManager asenseManager;
        asenseManager.initilize();
        asenseManager.run();
    }
    catch ( std::exception& ex ){
        std::cout << ex.what() << std::endl;
    }
}
